# -*- coding: utf-8 -*-

"""
Routines which make it easier to connect application in 
console application/script case and some desktop cases. 
"""

import webbrowser
import keyring
from rtm_connector import RtmConnector, RtmException, RtmServiceException
import getpass

def _console_invite_to_auth(): 
    print "Use the page just opened in the browser to let me access your RememberTheMilk account"

def _console_wait_for_user():
    raw_input("Press Enter once you finished browser authentication")

def create_and_authorize_connector(
    app_name, api_key, shared_secret,
    permission = "delete",
    invite_to_authorize = _console_invite_to_auth,
    wait_for_authorization = _console_wait_for_user):
    """
    Creates and returns RtmConnector object. Ensures it contains
    valid access token (if not, performs browser-based authorization).
    Grabbed token (if any) is saved in keyring for future use (using app_name
    to distnguish usage cases)
    """
    TOKEN_LABEL = "default-user-token"

    token = keyring.get_password(app_name, TOKEN_LABEL)

    connector = RtmConnector(api_key, shared_secret, permission, token)
    
    if not connector.token_valid():
        url, frob = connector.authenticate_desktop()
        invite_to_authorize()
        webbrowser.open(url)
        wait_for_authorization()
        if connector.retrieve_token(frob):
            keyring.set_password(app_name, TOKEN_LABEL, connector.token)
            print "Access token saved"
        else:
            raise RtmServiceException("Failed to grab access token")

    return connector

def _prompt_secure_data(what):
    return getpass.getpass("Enter %s: " % what)

def create_and_authorize_connector_prompting_for_api_key(
    app_name,
    permission = "delete",
    invite_to_authorize = _console_invite_to_auth,
    wait_for_authorization = _console_wait_for_user,
    prompt_secure_data = _prompt_secure_data):
    """
    Creates and authorizes connector interactively asing
    for api key (and saving it for future uses)
    """
    API_LABEL = "api-key"
    SEC_LABEL = "sec-code"

    api_key = keyring.get_password(app_name, API_LABEL)
    if not api_key:
        api_key = prompt_secure_data("API Key")
        keyring.set_password(app_name, API_LABEL, api_key)

    api_sec = keyring.get_password(app_name, SEC_LABEL)
    if not api_sec:
        api_sec = prompt_secure_data("shared secret")
        keyring.set_password(app_name, SEC_LABEL, api_sec)

    try:
        connector = create_and_authorize_connector(app_name, api_key, api_sec)
        return connector
    except RtmServiceException, e:
        if "Invalid API Key" in str(e):
            keyring.set_password(app_name, API_LABEL, "")
        if "Invalid signature" in str(e):
            keyring.set_password(app_name, SEC_LABEL, "")
        raise




# -*- coding: utf-8 -*-

import codecs
import csv

def csv_export(rtm_client, csv_filename, filter = None):
    raise Exception("Sorry, this is not yet implemented")
    output = codecs.open(csv_filename, "w", encoding = "utf-8")  
    csv_output = csv.writer(output)
    csv_output.writerow([u'Task', u'List', u'Recurring'])
    output.close()



# -*- coding: utf-8 -*-

usage = """
Simple RememberTheMilk data exporter.

Exporting data (lists and tasks) to JSON file:

    %prog --json=myrtm.json

   (in this case data are saved in the same format in which
    RememberTheMilk API provides them, just wrapped in separate
    fields for lists and tasks)

Exporting data (tasks) to CSV file:

    %prog --csv=myrtm.csv

Exported data can be narrowed using "--filter" clause, for example

    %prog --json=myrtm.json --filter="tag:Work and status:incomplete"

If filter is not set, "status:incomplete" is assumed.

"""

from connect import create_rtm_client

def parse_options():
    from optparse import OptionParser
    opt_parser = OptionParser(usage=usage)
    opt_parser.add_option("-j", "--json",
                          action="store", type="string", dest="json",
                          help="The name of .json file being exported")
    opt_parser.add_option("-c", "--csv",
                          action="store", type="string", dest="csv",
                          help="The name of .csv file being exported")
    opt_parser.add_option("-f", "--filter",
                          action="store", type="string", dest="filter",
                          help="Data filter (RTM query syntax)")
    opt_parser.add_option("-v", "--verbose",
                          action="store_true", dest="verbose",
                          help="Print diagnostic messages")
                          
    opt_parser.set_defaults(verbose = False)
    opt_parser.set_defaults(filter = "status:incomplete")
    (opts, args) = opt_parser.parse_args()

    if not (opts.json or opts.csv):
        opt_parser.error("Operation not selected (--json=filename or --csv=filename expected)")
    return opts

def run():

    opts = parse_options()
    client = create_rtm_client()

    if opts.csv:
        from mekk.rtm.exporters.csv import csv_export
        csv_export(client, opts.csv, opts.filter)
    if opts.json:
        from mekk.rtm.exporters.json import json_export
        json_export(client, opts.json, opts.filter)




# -*- coding: utf-8 -*-

usage = """
Simple RememberTheMilk importer.

Importing RTM export file (obtained via rtmexport --json=<file>):

    %prog --json=rtm.json  [--verbose] [--dry-run]

Importing Nozbe export file (obtained via nozbetool export --json=<file>):

    %prog --nozbe-json=nozbe.json  [--verbose] [--dry-run]
"""

from connect import create_rtm_client
import simplejson

def parse_options():
    from optparse import OptionParser
    opt_parser = OptionParser(usage=usage)
    opt_parser.add_option("-j", "--json",
                          action="store", type="string", dest="json",
                          help="The name of .json file exported using rtmexport --json ")
    opt_parser.add_option("-n", "--nozbe-json",
                          action="store", type="string", dest="nozbe_json",
                          help="The name of Nozbe .json file exported using nozbetool export --json ")
    opt_parser.add_option("-v", "--verbose",
                          action="store_true", dest="verbose",
                          help="Print diagnostic messages")
    opt_parser.add_option("-d", "--dry-run",
                          action="store_true", dest="dry_run",
                          help="Don't execute anything, just check input and (if verbose) print planned actions")
                          
    opt_parser.set_defaults(verbose = False, dry_run = False)
    (opts, args) = opt_parser.parse_args()

    if not (opts.json or opts.nozbe_json):
        opt_parser.error("Operation not selected (--json or --nozbe-json expected)")

    return opts

    
def run():

    opts = parse_options()
    client = create_rtm_client()

    if opts.json:
        data = simplejson.load(open(opts.json),
                               encoding = "utf-8")

        from mekk.rtm.importers.rtmjson import import_rtm

        import_rtm(
            client, tasks = data['tasks'], lists = data['lists'],
            verbose = opts.verbose, dry_run = opts.dry_run)

    if opts.nozbe_json:
        data = simplejson.load(open(opts.nozbe_json),
                               encoding = "utf-8")

        from mekk.rtm.importers.nozbe import import_nozbe_actions

        import_nozbe_actions(
            client, actions = data['actions'], notes = data['notes'],
            verbose = opts.verbose, dry_run = opts.dry_run)



from __access__ import API, SEC

from mekk.rtm.connect import create_and_authorize_connector
from mekk.rtm.rtm_client import RtmClient
from base64 import decodestring as __

def create_rtm_client(app_name = "mekk-rtm", permission = "delete"):
    """
    Configures and creates connected and authorized RtmClient.
    May use browser authorization to grab necessary permissions.
    Uses embedded API key, so appropriate only for scripts embedded
    in library!
    """
    connector = create_and_authorize_connector(
        app_name, __(API), __(SEC), permission)
    client = RtmClient(connector)
    return client



# -*- coding: utf-8 -*-

import codecs
import simplejson

def json_export(rtm_client, json_filename, filter = None):
    """
    Exports all task or task satisfying the filter.
    """
    raise Exception("Sorry, this is not yet implemented")

    lists = []
    tasks = []

    output = codecs.open(json_filename, "w", encoding = "utf-8")   
    simplejson.dump(
        dict(lists = lists, tasks = tasks),
        output,
        sort_keys = True,
        indent = 4,
        )
    output.close()
    



# -*- coding: utf-8 -*-

"""
Low-level RememberTheMilk connection handling.

Heavily inspired by rtmapi, but modified to use JSON instead of
(frequently confusing) XML, add logging, and be explicit
"""

import hashlib
import httplib2
import urllib
import simplejson
import logging

__all__ = ('RtmException', 'RtmConnector')

logger = logging.getLogger(__name__)

class RtmException(Exception): pass
class RtmConnectException(RtmException): pass
class RtmServiceException(RtmException): pass

class RtmConnector(object):
    _auth_url = "http://api.rememberthemilk.com/services/auth/"
    _base_url = "http://api.rememberthemilk.com/services/rest/"
    
    """
    @param api_key: your API key
    @param shared_secret: your shared secret
    @param perms: desired access permissions, one of "read", "write"
                  and "delete"
    @param token: token for granted access (optional, if not set,
                  authentication must be performed later)
    """
    def __init__(self, api_key, shared_secret, perms = "read", token = None):
        self.api_key = api_key
        self.shared_secret = shared_secret
        self.perms = perms
        self.token = token
        self.user_id = self.user_name = self.user_full_name = None
        self.http = httplib2.Http()
    
    ######################################################################
    # Authentication (routines to call if token is not yet known)
    ######################################################################

    """
    Authenticate as a desktop application.
    
    @returns: (url, frob) tuple with url being the url the user should open and
                          frob the identifier for usage with retrieve_token
                          after the user authorized the application
    """
    def authenticate_desktop(self):
        rsp = self.call_anonymously("rtm.auth.getFrob")
        frob = rsp["frob"]
        url = self._make_request_url(
            called_url = self._auth_url, perms=self.perms, frob=frob)
        return url, frob
    
    """
    Authenticate as a web application.
    @returns: url
    """
    def authenticate_webapp(self):
        url = self._make_request_url(
            called_url = self._auth_url, perms=self.perms)
        return url
    
    """
    Checks whether the stored token is valid. Apart from testing that,
    on success sets user_id, user_name and user_full_name

    @returns: bool validity
    """
    def token_valid(self):
        if self.token is None:
            return False
        try:
            rsp = self.call_anonymously("rtm.auth.checkToken",
                                        auth_token=self.token)
            self._set_user_attributes(rsp['auth']['user'])
        except RtmServiceException, e:
            logger.info("Failure verifying token: %s" % str(e))
            return False
        return True
    
    """
    Retrieves a token for the given frob

    Additionally, on success sets .token, .user_id, .user_name .user_full_name attributes

    @returns: True/False
    """
    def retrieve_token(self, frob):
        try:
            rsp = self.call_anonymously(
                "rtm.auth.getToken", frob=frob)
        except RtmServiceException, e:
            logger.warn("Failure retrieving token: %s" % str(e))
            self.token = None
            self.user_id = self.user_name = self.user_full_name = None
            return False
        self.token = rsp['auth']['token']
        self._set_user_attributes(rsp['auth']['user'])
        return True

    ######################################################################
    # Main call routines
    ######################################################################
    
    def call_anonymously(self, method_name, **params):
        """
        Executes given method without authorization credentials.

        @param method_name called metod (for example "rtm.auth.checkToken"
               or "rtm.test.echo")
        @param params any extra params named according to docs (note: params with value
               None will be dropped)
        @result output structure without wrappers

        See http://www.rememberthemilk.com/services/api/methods/ for method list.
        """
        return self._make_request(method = method_name, **params)
    
    def call_authorized(self, method_name, **params):
        """
        Executes given method with authorization credentials.

        @param method_name called metod (for example "rtm.auth.checkToken"
               or "rtm.test.echo")
        @param params any extra params named according to docs (note: params with value
               None will be dropped)
        @result output structure without wrappers

        See http://www.rememberthemilk.com/services/api/methods/ for method list.
        """
        return self._make_request(method = method_name, auth_token = self.token, **params)

    call = call_authorized
    
    ######################################################################
    # Internal helpers
    ######################################################################

    def _make_request(self, method, called_url = None, **params):
        final_url = self._make_request_url(called_url = called_url, method=method, **params)
        logger.debug("Calling %s" % final_url)
        info, data = self.http.request(
            final_url,
            headers={'Cache-Control':'no-cache, max-age=0'})
        if info['status'] != '200':
            raise RtmConnectException("Connection failure: %s (%s)" % (
                    info['status'], info.get('reason', '')))
        logger.debug("Call to %s executed. Result: %s" % (method, data))
        reply = simplejson.loads(data)
        rsp = reply['rsp']
        if rsp['stat'] != "ok":
            err = rsp.get("err")
            if err:
                raise RtmServiceException(
                    "Service failure: %(code)s (%(msg)s)" % err)
            else:
                raise RtmServiceException(
                    "Unknown service failure during call to %s" % final_url)
        return rsp
    
    def _make_request_url(self, called_url = None, **params):
        """
        Construct final REST-style URL. Adds format, api_key and signature.
        """
        params["api_key"] = self.api_key
        params["format"] = "json"
        param_pairs = [ (k,v) for k,v in params.iteritems() 
                              if v is not None ]
        param_pairs.append( ("api_sig", self._sign_request(param_pairs)) )
        quote_utf8 = lambda s: urllib.quote_plus(s.encode('utf-8'))
        params_joined = "&".join("%s=%s" % (quote_utf8(k), quote_utf8(v))
                                           for k, v in param_pairs)
        return (called_url or self._base_url) + "?" + params_joined
    
    def _sign_request(self, param_pairs):
        param_pairs.sort()
        request_string = self.shared_secret + u''.join(k+v
                                                       for k, v in param_pairs)
        return hashlib.md5(request_string.encode('utf-8')).hexdigest()

    def _set_user_attributes(self, user):
        self.user_id = user['id']
        self.user_name = user['username']
        self.user_full_name = user['fullname']




from mekk.rtm.rtm_connector import RtmConnector, RtmException, RtmServiceException, RtmConnectException
from mekk.rtm.rtm_client import RtmClient, RtmDataException
from mekk.rtm.connect import create_and_authorize_connector_prompting_for_api_key



# -*- coding: utf-8 -*-

usage = r"""
Move all tags satisfying given criteria to given list.

Example:

    %prog --list='Language learning' \
            --filter='list:"English Course" or tag:"French" or tag:German"

Destination list is created if not yet present.

Note: filter is using standard RememberTheMilk language, see
http://www.rememberthemilk.com/help/answers/search/advanced.rtm
"""

from connect import create_rtm_client

def parse_options():
    from optparse import OptionParser
    opt_parser = OptionParser(usage=usage)
    opt_parser.add_option("-l", "--list",
                          action="store", type="string", dest="list",
                          help="Target list to which all specified tasks will be moved.")
    opt_parser.add_option("-f", "--filter",
                          action="store", type="string", dest="filter",
                          help="Data filter (RTM query syntax)")
    opt_parser.add_option("-v", "--verbose",
                          action="store_true", dest="verbose",
                          help="Print diagnostic messages")
    opt_parser.add_option("-d", "--dry-run",
                          action="store_true", dest="dry_run",
                          help="Don't execute any updates, just check input and (if verbose) print planned actions")
                          
    opt_parser.set_defaults(verbose = False, dry_run = False, add =[], remove =[])
    (opts, args) = opt_parser.parse_args()

    if not opts.filter:
        opt_parser.error("--filter not specified")
    if not opts.list:
        opt_parser.error("--list not specified")
    return opts
    
def run():

    opts = parse_options()
    client = create_rtm_client()

    raise Exception("Sorry, this is not yet implemented")

    print opts.add
    print opts.remove
    print opts.filter

# ...




# -*- coding: utf-8 -*-

from collections import namedtuple
from dateutil.parser import parse as dateutil_parse
from dateutil.tz import tzutc
from mekk.rtm.rtm_connector import RtmException
import datetime

############################################################
# Helper structures (representations of RTM data)
############################################################

"""
See RtmClient class for API documentation.

The following structures (namedtuples) are used to represent
data downloaded from RememberTheMilk

List (normal list):
- id (string) - unique id
- name (string) - list name
- archived (bool) - is the list archived (true) or active (false)
"""
List = namedtuple('List', 'id name archived')  

"""
Smart list (query):
- id, name, archived - as in List
- filter (string) - smartlist query (as defined via web interface)
"""
SmartList = namedtuple('SmartList', 'id name filter archived')

"""
Note (extra text appended to task)
- id (string) - unique id (inside task)
- title (string) - note title
- body (string) - note text (can contain newlines)
"""
Note = namedtuple('Note', 'id title body')

"""
Unique identification of individual task:
- list_id (string) - identification of list the task belons to
- taskseries_id (string) - ...
- task_id (string) - id of task as such
"""
TaskKey = namedtuple('TaskKey', 'list_id task_id taskseries_id')

"""
All task data:
- key (TaskKey) - task identification
- name (string) - task description
- tags (list of strings, can be empty) - all task tags
- notes (list of Note objects, can be empty) - all notes appended to task
- due (datetime.datetime or None) - due date, can be None when not set
- estimate (string or None) - estimated cost in RTM jargon (f.e. "5 hours"), None when unset
- priority (int or None) - priority (1,2,3 or None when unset)
- completed (bool) - is the task already completed?
- postponed (int) - how many times task was postponed (0 if never)
- repeat (string or None) - textual description of repeating clause
     (currently sth like "every FREQ=DAILY;INTERVAL=3", in the future
      can change to sth more sensible)
- url (string or None) - url bound to task if any
- deleted (bool) - set to true if the task has been deleted (can happen only
      in result of delete method)
"""
Task = namedtuple('Task', 'key name tags notes due estimate priority completed postponed repeat url deleted')

class RtmDataException(RtmException): 
    """
    Exception thrown when the data structure is not understood by the library.
    """
    pass


# Implementation note: I abstract TaskSeries->Task hierarchy, using
# only Tasks (and adding taskseries as attribute to them). While taskseries
# help aggregate some attributes in case of repeating tasks, I don't feel
# they are of much use from the client point of view.

class RtmClient(object):
    """
    Wrapper for RTM client calls. Handles listing, creating
    and modifying lists, tags, locations, and tasks.

    Apart from wrapping RememberTheMilk API quirks it:

    - keeps the cache of known lists to avoid re-downloading them and
      handles duplicate detection in case of lists (so the same list
      is not created again)

    - abstracts over some differences in data structure and format
      providing uniform API.
      
    All updates are performed inside single timeline (Rtm undo
    context) unless set_undo_point routine is called.
    """
    
    def __init__(self, connector):
        """
        Initializes object. 

        @param connector authorized mekk.rtm.RtmConnector object used to
               handle communication. Usually created using
               functions from mekk.rtm.connect module
        """
        self.connector = connector
        self._list_cache = None      # name -> List
        self._smartlist_cache = None # name -> SmartList
        self._timeline = None

    def find_or_create_list(self, list_name):
        """
        Looks for the (normal) list of given name.
        If such list does not exist, creates it.

        Uses list cache for better performance.

        Returns List object representing given list
        """
        self._load_list_cache_if_necessary()
        list_info = self._list_cache.get(list_name)
        if list_info:
            #if list_info.archived:
            #    self.unarchive_list(list_info.id)
            return list_info
        timeline = self._get_timeline()

        r = self.connector.call(
            "rtm.lists.add",
            timeline = timeline,
            name = list_name)
        new_list = self._process_list_reply(r['list'])
        return new_list

    def unarchive_list(self, list_id):
        """
        Unarchives list of given id, returns the List or SmartList
        object representing list state after the operation.
        """
        r = self.connector.call(
            "rtm.lists.unarchive", 
            timeline = self._get_timeline(),
            list_id = list_id)
        return self._process_list_reply(r['list'])

    def archive_list(self, list_id):
        """
        Archives list of given id, returns the List or SmartList
        object representing list state after the operation.
        """
        r = self.connector.call(
            "rtm.lists.archive",
            timeline = self._get_timeline(),
            list_id = list_id)
        return self._process_list_reply(r['list'])

    def delete_list(self, list_id):
        """
        Deletes list of given id. Returns List object showing before-deletion data
        """
        r = self.connector.call(
            "rtm.lists.delete",
            timeline = self._get_timeline(),
            list_id = list_id)
        return self._process_list_reply(r['list'])

    def known_lists(self):
        """
        Returns all known lists (except smartlists, only true lists are returned).

        Return value is a list of List namedtuples
        """
        self._load_list_cache_if_necessary()
        return self._list_cache.values()

    def known_smart_lists(self):
        """
        Returns all known smart lists.

        Return value is a list of SmartList namedtuples
        """
        self._load_list_cache_if_necessary()
        return self._smartlist_cache.values()

    def create_task(self, task_name,
                    list_id = None,
                    tags = [], 
                    completed = False,
                    priority = None,
                    due_date = None,
                    estimate = None,
                    repeat = None,
                    url = None,
                    notes = [],
                    ):
        """
        Creates new task and sets some attributes. Returns TaskKey
        object needed to update the task later.

        @param task_name 
        @param list_id list identifier, can be None (then Inbox is used)
        @param tags list of tags (note: tags can't contain ',')
        @param completed should task be immediately marked as completed?
        @param priority  1, 2, or 3 (or None)
        @param due_date due date (datetime.date or ISO 8601 text, if given date is zone-unaware, it is assumed to be UTC)
        @param estimate estimated cost (text like "3 days 2 hours 40 minutes")
        @param repeat task repeating clause (text like "Every Tuesday", "Every month on the 4th", "Every week until 1/1/2007", "After a week" etc. See http://www.rememberthemilk.com/help/answers/basics/repeatformat.rtm )
        @param url webpage related to the task (or None)
        @param notes list of pairs (title, text) representing notes to be added to the task

        @returns Task object representing the task
        """
        timeline = self._get_timeline()
        r = self.connector.call("rtm.tasks.add",
                                timeline = timeline,
                                list_id = list_id,
                                name = task_name)
        if not list_id:
            # addTags and other methods require it
            list_id = r['list']['id']
        task = RtmClient._parse_task(list_id, RtmClient._ensure_element_is_singular(r['list']['taskseries']))

        self.update_task(task.key,
                         add_tags = tags,
                         completed = completed,
                         priority = priority,
                         due_date = due_date,
                         estimate = estimate,
                         repeat = repeat,
                         url = url,
                         notes = notes,
                         )

        return task

    def update_task(self, task_key,
                    add_tags = [],  # add tags without changing existing
                    set_tags = None, # If specified, list of tags, drop old tags
                    completed = False,
                    priority = None,
                    due_date = None,
                    estimate = None,
                    repeat = None,
                    url = None,
                    notes = [],
                    ):
        """
        Updates given task attributes. All parameters has
        the same semantic as in create_task except  set_tags,
        which, if specified, resets tag list to given items
        (removing all old tags). 
        """
        timeline = self._get_timeline()

        r = None

        if set_tags is not None:
            tag_text = ", ".join(set_tags)
            r = self.connector.call(
                "rtm.tasks.setTags",
                timeline = timeline,
                tags = tag_text,
                **task_key._asdict())            
            
        if add_tags:
            tag_text = ", ".join(add_tags)
            r = self.connector.call(
                "rtm.tasks.addTags",
                timeline = timeline,
                tags = tag_text,
                **task_key._asdict())

        if completed:
            r = self.connector.call(
                "rtm.tasks.complete",
                timeline = timeline,
                **task_key._asdict())

        if priority:
            r = self.connector.call(
                "rtm.tasks.setPriority",
                timeline = timeline,
                priority = str(priority),
                **task_key._asdict())

        if due_date:
            r = self.connector.call(
                "rtm.tasks.setDueDate",
                timeline = timeline,
                due = RtmClient._prepare_date_for_sending(due_date),
                parse = "1", # has_due_time = 1
                **task_key._asdict())

        if estimate:
            r = self.connector.call(
                "rtm.tasks.setEstimate",
                timeline = timeline,
                estimate = estimate,
                **task_key._asdict())

        if repeat:
            r = self.connector.call(
                "rtm.tasks.setRecurrence",
                timeline = timeline,
                repeat = repeat,
                **task_key._asdict())
            
        if url:
            r = self.connector.call(
                "rtm.tasks.setURL",
                timeline = timeline,
                url = url,
                **task_key._asdict())

        if r:
            task = RtmClient._parse_task(task_key.list_id,
                                         RtmClient._ensure_element_is_singular(r['list']['taskseries']))
        else:
            task = None

        for (note_title, note_text) in notes:
            rn = self.add_task_note(task_key,
                               note_title = note_title, note_text = note_text)
            if task:
                task.notes.append(rn)

        return task

    def add_task_note(self, task_key, 
                      note_title, note_text):
        """
        Adds new note to the task. 

        @param task_key: task identity
        @param note_title: note label (invisible?)
        @param note_text: actual body
        @returns: Note object repesenting note
        """
        timeline = self._get_timeline()
        r = self.connector.call(
            "rtm.tasks.notes.add",
            timeline = timeline,
            note_title = note_title, note_text = note_text,
            **task_key._asdict())
        return RtmClient._parse_note(r['note'])

    def delete_task(self, task_key):
        """
        Deletes given task.
        @returns: Task data after operation (with deleted=True)
        """
        timeline = self._get_timeline()
        r = self.connector.call(
            "rtm.tasks.delete",
            timeline = timeline,
            **task_key._asdict())
        return RtmClient._parse_task(task_key.list_id,
                                     RtmClient._ensure_element_is_singular(r['list']['taskseries']))

    def find_tasks(self,
                   list_id = None, 
                   filter = None,
                   last_sync = None):
        """
        Finds all asks matchin given criteria: belonging to specified
        list (or any list if missing), satisfying filter if present,
        and modified since last_sync (if present). last_sync can be
        specified as datetime

        Filter syntax: http://www.rememberthemilk.com/help/answers/search/advanced.rtm

        The routine works as generator and yields all found tasks. To
        get the results as list, just wrap it with list(), for example::
        
            my_items = list(client.find_tasks(list_id = inbox.id))
        """
        r = self.connector.call(
            "rtm.tasks.getList",
            list_id = list_id,
            filter = filter,
            last_sync = last_sync)
        tasks = r['tasks'].get('list', [])
        if type(tasks) is dict:
            tasks = [ tasks ]
        for task_list in tasks:
            list_id = task_list['id']
            for taskseries in RtmClient._wrap_element_as_list_if_singular(task_list['taskseries']):
                task_block = taskseries['task']
                if type(task_block) is not list:
                    task_block = [ task_block ]
                for task in task_block:
                    yield RtmClient._parse_task(list_id, taskseries, task)

    def set_undo_point(self):
        """
        Force creation of new timeline. Timelines are not really used 
        in the library yet, so it is more marker of possible future 
        development than anything else.
        """
        # Just forgetting previous timeline. New one will be created before first update
        self._timeline = None

    def _get_timeline(self):
        if self._timeline is None:
            r = self.connector.call(
                "rtm.timelines.create")
            self._timeline = r['timeline']
        return self._timeline

    def _load_list_cache_if_necessary(self):
        if self._list_cache is None:
            r = self.connector.call(
                "rtm.lists.getList")
            self._list_cache = {}
            self._smartlist_cache = {}
            for l in r['lists']['list']:
                self._process_list_reply(l)

    def _process_list_reply(self, list_dic):
        """
        Converts XML list data (as returned by many rtm.lists commands)
        to List or SmartList tuple and updates internal cache for this list.
        """
        id = list_dic['id']
        name = list_dic['name']
        archived = RtmClient._parse_flag(list_dic['archived'])
        deleted = RtmClient._parse_flag(list_dic['deleted'])
        if not RtmClient._parse_flag(list_dic['smart']):
            r = List(
                id = id,
                name = name,
                archived = archived)
            if not deleted:
                self._list_cache[name] = r
            else:
                del self._list_cache[name]
        else:
            r = SmartList(
                id = id,
                name = name,
                archived = archived,
                filter = list_dic['filter'])
            if not deleted:
                self._smartlist_cache[name] = r
            else:
                del self._smartlist_cache[name]
        return r

    @staticmethod
    def _parse_task(list_id, taskseries_blk, task_blk = None):
        """
        Converts XML data to Task. If task_blk is not given, assumes first
        """
        if task_blk is None:
            task_blk = taskseries_blk['task']

        tags = RtmClient._unwrap_optional_element(taskseries_blk['tags'], 'tag', [])
        notes = RtmClient._unwrap_optional_element(taskseries_blk['notes'], 'note', [])

        return Task(
            key = TaskKey(list_id = list_id, taskseries_id = taskseries_blk['id'], task_id = task_blk['id']),
            name = taskseries_blk['name'],
            tags = tags,
            notes = [ RtmClient._parse_note(n) for n in notes ],
            due = RtmClient._parse_due(task_blk['due']),
            estimate = RtmClient._parse_estimate(task_blk['estimate']),
            priority = RtmClient._parse_priority(task_blk['priority']),
            completed = RtmClient._parse_completed(task_blk['completed']),
            postponed = RtmClient._parse_postponed(task_blk['postponed']),
            repeat = RtmClient._parse_rrule(taskseries_blk.get('rrule')),
            url = RtmClient._parse_url(taskseries_blk.get('url', '')),
            deleted = RtmClient._parse_deleted(task_blk['deleted']),
            )

    @staticmethod
    def _parse_date(date_attr):
        if date_attr:
            return dateutil_parse(date_attr)
        else:
            return None

    _parse_due = _parse_date
    _parse_completed = _parse_date
    _parse_deleted = _parse_date

    @staticmethod
    def _parse_flag(flag_text):
        """
        Parses logical flag set as "1" or "0" (and sometimes ""). Returns True/False
        """
        return bool(int(flag_text or "0"))
        
    @staticmethod
    def _parse_postponed(txt):
        return int(txt)

    @staticmethod
    def _parse_txt_treating_empty_as_none(txt):
        if txt:
            return txt
        else:
            return None

    _parse_estimate = _parse_txt_treating_empty_as_none
    _parse_url = _parse_txt_treating_empty_as_none

    @staticmethod
    def _parse_priority(txt):
        if txt and txt != "N":
            return int(txt)
        else:
            return None

    @staticmethod
    def _parse_rrule(rrule_blk):
        # TODO: convert those ugly FREQ=DAILY;INTERVAL=1 back to 1 day etc
        if not rrule_blk:
            return None
        if rrule_blk.get('after'):
            pfx = "after"
        elif rrule_blk.get('every'):
            pfx = "every"
        return pfx + " " + rrule_blk['$t']

    @staticmethod
    def _parse_note(note_blk):
        return Note(id = note_blk['id'],
                    title = note_blk['title'],
                    body = note_blk['$t'])
            
    @staticmethod
    def _prepare_date_for_sending(some_date):
        """
        Normalizes date before posting to RTM server: flags it as UTC if it is zoneless,
        converts to string if it is datetime.date or datetime.datetime.
        """
        if type(some_date) is datetime.date:
            return some_date.isoformat() + "T00:00:00Z"
        if not ((type(some_date) is datetime.datetime)):
            some_date = dateutil_parse(some_date)
        if not some_date.tzinfo:
            some_date = some_date.replace(tzinfo=tzutc())
        return some_date.isoformat()

    @staticmethod
    def _unwrap_optional_element(item, tag_name, default_value=[]):
        """
        Handles the case when given element either contains empty list, or subdictionary with
        interesting data pointed by single key. Namely the case of tags/notes which can
        look so:

           "tags": []

        or so:

           "tags":{"tag":["@dom","testy"]}
        
        @param item analyzed element in the tree (for the above example, either [] or {"tag":["@dom","testy"]})
        @param tag_name name of the sub-item
        @return value of the subitem, or default value (used when empty list is provided)
        """
        if type(item) is list:
            if item != []:
                raise RtmDataException("Block expected to be empty list or dictionary is nonempty list: %s\nNotify developer" % str(item))
            return default_value
        else:
            return item[tag_name]

    @staticmethod
    def _wrap_element_as_list_if_singular(item):
        """
        Handles the case, when some element either is a list of items (usually list of dictionaries),
        or single item. Example case is that of taskseries, which happen to be:

           "taskseries":[
              {"id":999, ... },
              {"id":888, ... },
           ]

        but also:

           "taskseries":{"id":999, ... }

        Always returns the list (for the latter case, consisting of single element).
        """
        if type(item) is list:
            return item
        else:
            return [item]

    @staticmethod
    def _ensure_element_is_singular(item):
        """
        Handles the same duality as _wrap_element_as_list_if_singular but expects
        exactly one item (be it on the list, or without it), otherwise raises exception
        """
        l = RtmClient._wrap_element_as_list_if_singular(item)
        if len(l) != 1:
            raise RtmDataException("Expected exactly one item, got less or more: %s\nNotify developer." % str(item))
        return l[0]
